/**
 * @copyright Copyright 2013 by the Websoft Research Group, Nanjing University, PR China
 * @license MIT (https://opensource.org/licenses/MIT)
 * @link http://ws.nju.edu.cn/ptime
 */
package cn.edu.nju.ws.ptime.core;

import cn.edu.nju.ws.ptime.tml.ExpressionIndex;
import cn.edu.nju.ws.ptime.tml.TDocument;
import cn.edu.nju.ws.ptime.tml.TaggedToken;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

public class ExpressionExtractor {
	private String[][] patterns;

	//TODO: The naive extraction can be optimized (e.g. use trie or automata), but since it is not the efficiency bottleneck ...

	public ExpressionExtractor(String[][] patterns) {
		this.patterns = patterns;
	}

	private List<ExpressionIndex> extractFromTDoc(int docId, TDocument tDoc, String[] pattern) {
		List<ExpressionIndex> timexList = new ArrayList<>();
		TaggedToken[] tokens = tDoc.taggedTokens();
		for (int i = 0; i < tokens.length; ++i) {
			int j = 0;
			while (j < pattern.length && i + j < tokens.length && tokens[i + j].tok2string().matches(pattern[j])) {
				j++;
			}
			if (j == pattern.length) {
				timexList.add(new ExpressionIndex(docId, i, i + j, "", ""));
			}
		}
		return timexList;
	}

	/**
	 * Extract time expression for measure the quality of patterns
	 * @param tDocs parsed documents
	 * @return extracted expressions (without any processing)
	 */
	public List<List<ExpressionIndex>> extractForMeasurement(TDocument[] tDocs) {
		return Arrays.stream(patterns).parallel().map(pattern -> {
			List<ExpressionIndex> timexList = new ArrayList<>();
			for (int k = 0; k < tDocs.length; ++k) {
				timexList.addAll(extractFromTDoc(k, tDocs[k], pattern));
			}
			return timexList;
		}).collect(Collectors.toList());
	}

	/**
	 * Recognize time expression from token sequence
	 * @param tDoc parsed document
	 * @return extracted expressions (without any processing)
	 */
	public List<ExpressionIndex> extractForRecognition(TDocument tDoc) {
		return Arrays.stream(patterns).parallel()
		             .flatMap(pattern -> extractFromTDoc(0, tDoc, pattern).stream())
		             .collect(Collectors.toList());
	}
}
