/**
  * @copyright Copyright 2013 by the Websoft Research Group, Nanjing University, PR China
  * @license MIT (https://opensource.org/licenses/MIT)
  * @link http://ws.nju.edu.cn/ptime
  */
package cn.edu.nju.ws.ptime.tml

case class ExpressionIndex(
 docId: Int,
 from: Int,
 until: Int,
 value: String,
 eType: String
) extends Comparable[ExpressionIndex] {
  def isContainedIn(x: ExpressionIndex): Boolean = {
    docId == x.docId && from >= x.from && until <= x.until
  }
  def length: Int = until - from

  def containedCoverage(x: ExpressionIndex): Double = {
    if (isContainedIn(x)) length.toDouble / x.length else 0.0
  }

  def overlap(x: ExpressionIndex): Boolean = {
    docId == x.docId && Math.max(from, x.from) < Math.min(until, x.until)
  }

  def overlapOrConsecutive(x: ExpressionIndex): Boolean = {
    docId == x.docId && Math.max(from, x.from) <= Math.min(until, x.until)
  }

  def before(x: ExpressionIndex): Boolean = {
    docId < x.docId || (docId == x.docId && until <= x.from)
  }
  def beforeOrOverlap(x: ExpressionIndex): Boolean = {
    docId < x.docId || (docId == x.docId && from < x.until)
  }

  override def compareTo(o: ExpressionIndex): Int = {
    if (docId != o.docId) docId - o.docId
    else if (from != o.from) from - o.from
    else until - o.until
  }
}
